
//
// cle4vdr 2005 - 2018 by Alexander Hans <cle4vdr@arcor.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
// Or, point your browser to http://www.gnu.org/copyleft/gpl.html
//


unit uchannels;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ComCtrls, StrUtils, Dialogs;

type TAudioArray = Array of Array[0..2] of String;

type
    cleP = ^cleChannels;
    cleChannels = Record
    Number: Integer;
    GroupSep: Boolean;
    Name : String;
    Short: String;
    Provider: String;
    Source: String;
    Frequency: String;
    Pol: String;
    Symbolrate: String;
    Videopid: String;
    VideoType: String;
    PcrPid: String;
    AudioPids: TAudioArray;
    AC3Pids: TAudioArray;
    Ttpid: String;
    SubtPids: TStringList;
    Ca: String;
    Sid: String;
    Nid: String;
    Tid: String;
    Rid: String;
         { Values taken from vdr.5 2.39
              B   Bandwidth (1712, 5, 6, 7, 8, 10)
	      C   Code rate high priority (0, 12, 23, 34, 35, 45, 56, 67, 78, 89, 910)
	      D   coDe rate low priority (0, 12, 23, 34, 35, 45, 56, 67, 78, 89, 910)
	      G   Guard interval (4, 8, 16, 32, 128, 19128, 19256)
	      H   Horizontal polarization
	      I   Inversion (0, 1)
	      L   Left circular polarization
	      M   Modulation (2, 5, 6, 7, 10, 11, 12, 16, 32, 64, 128, 256, 999)
	      N   pilot mode (0, 1, 999)
	      O   rollOff (0, 20, 25, 35)
	      P   stream id (0-255)
	      Q   t2 system id (0-65535)
	      R   Right circular polarization
	      S   delivery System (0, 1)
	      T   Transmission mode (1, 2, 4, 8, 16, 32)
	      V   Vertical polarization
	      X   siso/miso mode (0, 1)
	      Y   hierarchY (0, 1, 2, 4)      }
    Bandwidth: String;
    CodeRH: String;
    CodeRL: String;
    Guard: String;
    Inversion: String;
    Modulation: String;
    PilotMode: String;
    RollOff: String;
    StreamID: String;    // (0-255)
    T2SysID: String;
    System: String;      // (0), DVB-S/T, (1) DVB-S2/T2
    TransmissionMode: String;
    SiMiMode: String;
    Hierarchy: String;
    IPTVParams: String;
  end;


var
  TVP: cleP;
  CopyData: array of cleP;
  BandwidthMap, CodeRMap, GuardMap, ModulationMap, PilotModeMap, HierarchyMap,
  InversionMap, RolloffMap, TransmissionMap,
  SystemMap, PolMap, SiMiModeMap: TStringList;

const
  NEWCHANNELSTR = 'New Channel:0:0:0:0:0:0:0:0:0:0:0:0';

procedure AssignNodeImage(n: TTreeNode);
procedure AddGroupSepStr(str: String);
function NewGroupSep(str: String): Pointer;
procedure AddChannelStr(str: String);
function NewChannel(str: String): Pointer;
procedure FillChannelData(str: String; p: cleP);
procedure InsertGroupSep(str: String);
function InsertGroupSep(t: TTreeNode; str: String): TTreeNode;
procedure InsertChannel(str: String; n: TTreeNode; update: Boolean);
procedure InsertNewChannel(str: String; n: TTreeNode);

function LongChannelName(str: String): String;
function ShortChannelName(str: String): String;
function ProviderName(str: String): String;

function GetProviderStr(n: TTreeNode): String;
function GetChannelStr(n: TTreeNode; old: Boolean=false): String;
function AudioPidsToStr(n: TTreeNode): String;

function isTV(n: TTreeNode): Boolean;
function isRadio(n: TTreeNode): Boolean;
function isUnknown(n: TTreeNode): Boolean;
function isGroupSep(n: TTreeNode): Boolean;
function isIPTV(n: TTreeNode): Boolean;
function isEncrypted(n: TTreeNode): Boolean;
function ChannelType(n: TTreeNode): String;
function ChanIDFreq(pol, freq: String): String;
function ChannelID(n: TTreeNode): String;
function LanguagesToStr(n: TTreeNode): String;
function HasDolby(n: TTreeNode): Boolean;
function HasAudio(n: TTreeNode): Boolean;
function HasMultiLang(n: TTreeNode): Boolean;

function GetAudioPid(str: String): String;
function GetLanguage(str: String): String;
function GetAudVidType(str: String): String;
function GetVidPid(str: String): String;
function GetVidType(str: String): String;
function GetPcrPid(str: String): String;
function GetSubtitlePids(str: String): TStringList;
function GetParam(str: String; c: Char): String;
procedure GetAudioPids(str: String; var AudArray: TAudioArray);
procedure GetAC3Pids(str: String; var Ac3Array: TAudioArray);
function PolToStr(str: String): String;
function EncryptionsToStr(n: TTreeNode): String;

procedure RenumberChannels;
function GetPrevChannel(n: TTreeNode): TTreeNode;
function GetNextChannel(n: TTreeNode): TTreeNode;
function GetNextGroupSep(n: TTreeNode): TTreeNode;
function HasWrongGroupSep(n: TTreeNode): Boolean;
function GetRightGroupSep(n: TTreeNode): TTreeNode;

function GetAllSources(var sl: TStringList): Integer;
procedure RemoveSources(source: String);
procedure RemoveAll(target: String);

implementation

uses uinterface, uutils, uinfo, uedit;

procedure BuildParameterMaps;
begin
  BandwidthMap    := TStringList.Create;
  CodeRMap        := TStringList.Create;
  GuardMap        := TStringList.Create;
  InversionMap    := TStringList.Create;
  ModulationMap   := TStringList.Create;
  PilotModeMap    := TStringList.Create;
  RolloffMap      := TStringList.Create;
  SystemMap       := TStringList.Create;
  TransmissionMap := TStringList.Create;
  SiMiModeMap     := TStringList.Create;
  HierarchyMap    := TStringList.Create;
  PolMap          := TStringList.Create;
  BandwidthMap.CommaText    := '1712=1712kHz,5=5MHz,6=6MHz,7=7MHz,8=8MHz,10=10MHz';
  CodeRMap.CommaText        := '0=0,12=1/2,23=2/3,34=3/4,35=3/5,45=4/5,56=5/6,67=6/7,78=7/8,89=8/9,910=9/10';
  GuardMap.CommaText        := '4=1/4,8=1/8,16=1/16,32=1/32,128=1/128,19128=19/128,19256=19/256';
  InversionMap.CommaText    := '0=0,1=1';
  ModulationMap.CommaText   := '2=QPSK,5=8PSK,6=16APSK,7=32APSK,10=VSB8,11=VSB16,12=DQPSK,16=QAM16,32=QAM32,64=QAM64,128=QAM128,256=QAM256';
  PilotModeMap.CommaText    := '0=off,1=on,999=auto';
  RolloffMap.CommaText      := '0=0,20=0.20,25=0.25,35=0.35';
  SystemMap.CommaText       := '0=DVB-S/T,1=DVB-S2/T2';
  TransmissionMap.CommaText := '1=1k,2=2k,4=4k,8=8k,16=16k,32=32k';
  SiMiModeMap.CommaText     := '0=SISO,1=MISO';
  HierarchyMap.CommaText    := '0=0,1=1,2=2,4=4';
  PolMap.CommaText          := 'H=Horizontal,V=Vertical,L=Left circular,R=Right circular';
end;

procedure RemoveParameterMaps;
begin
  BandwidthMap.Free;
  CodeRMap.Free;
  GuardMap.Free;
  InversionMap.Free;
  ModulationMap.Free;
  PilotModeMap.Free;
  RolloffMap.Free;
  SystemMap.Free;
  TransmissionMap.Free;
  HierarchyMap.Free;
  PolMap.Free;
end;

procedure AssignNodeImage(n: TTreeNode);
begin
  if n = NIL then Exit;
  if isGroupSep(n) then n.ImageIndex := 0
  else if isTV(n) then
  begin
    if isEncrypted(n) then n.ImageIndex := 2
    else n.ImageIndex := 1;
  end
  else if isRadio(n) then
  begin
    if isEncrypted(n) then n.ImageIndex := 4
    else n.ImageIndex := 3;
  end
  else n.ImageIndex := 5;
  n.SelectedIndex := n.ImageIndex;
end;


procedure AddGroupSepStr(str: String);
begin
  if length(str) < 1 then Exit;
  str := copy(str, 2, length(str));
  str := AnsiReplaceStr(str, '|', ':');
  Node := frmMain.tvMain.Items.GetLastNode;
  Node := frmMain.tvMain.Items.AddObject(Node, str, NewGroupSep(str));
  AssignNodeImage(Node);
end;

function NewGroupSep(str: String): Pointer;
begin
  New(TVP);
  result           := TVP;
  TVP^.Number      := 0;
  TVP^.GroupSep    := true;
  TVP^.Name        := str;
  TVP^.Short       := '';
  TVP^.Provider    := '';
  TVP^.Source      := '';
  TVP^.Frequency   := '';
  TVP^.Pol         := '';
  TVP^.Symbolrate  := '';
  TVP^.Videopid    := '';
  TVP^.VideoType   := '';
  TVP^.PcrPid      := '';
  TVP^.AudioPids   := NIL;
  TVP^.AC3Pids     := NIL;
  TVP^.TTpid       := '';
  TVP^.SubtPids    := NIL;
  TVP^.Ca          := '';
  TVP^.Sid         := '';
  TVP^.Nid         := '';
  TVP^.Tid         := '';
  TVP^.Rid         := '';
  TVP^.CodeRH      := '';
  TVP^.CodeRL      := '';
  TVP^.Guard       := '';
  TVP^.Modulation  := '';
  TVP^.StreamID    := '';
  TVP^.Hierarchy   := '';
  TVP^.Bandwidth   := '';
  TVP^.Inversion   := '';
  TVP^.RollOff     := '';
  TVP^.PilotMode   := '';
  TVP^.SiMiMode    := '';
  TVP^.TransmissionMode := '';
  TVP^.System      := '';
  TVP^.IPTVParams  := '';
end;

procedure AddChannelStr(str: String);
var
  dummy: String;
  n: TTreeNode;
begin
  dummy := str;
  str := copy(str, 1, AnsiPos(':', str)-1);
  str := AnsiReplaceStr(str, '|', ':');
  if (frmMain.tvMain.Items.Count > 0) AND
     ((isGroupSep(frmMain.tvMain.Items.GetLastNode)) OR (frmMain.tvMain.Items.GetLastNode.Parent <> NIL)) then
    n := frmMain.tvMain.Items.AddChildObject(Node, LongChannelName(str), NewChannel(dummy))
  else
    n := frmMain.tvMain.Items.AddObject(Node, LongChannelName(str), NewChannel(dummy));
  AssignNodeImage(n);
end;

function NewChannel(str: String): Pointer;
begin
  New(TVP);
  result := TVP;
  FillChannelData(str, TVP);
end;

procedure FillChannelData(str: String; p: cleP);
var
  strl: TStringList;
begin
  strl := StringSplit(str, ':');
  if strl.Count > 10 then
  begin
    p^.Number     := 0;
    p^.GroupSep   := false;
    strl[0] := AnsiReplaceStr(strl[0], '|', ':');
    p^.Name       := LongChannelName(strl[0]);
    p^.Short      := ShortChannelName(strl[0]);
    p^.Provider   := ProviderName(strl[0]);
    p^.Frequency  := strl[1];
    if strl[2] <> '' then
    begin
      if (UpperCase(strl[3]) = 'P') OR (UpperCase(strl[3]) = 'I') then p^.Pol := strl[2]
      else p^.Pol := strl[2][1]
    end
    else p^.Pol := '';
    p^.Source     := UpperCase(strl[3]);
    p^.Symbolrate := strl[4];
    p^.Videopid   := GetVidPid(strl[5]);
    p^.VideoType  := GetVidType(strl[5]);
    p^.PcrPid     := GetPcrPid(strl[5]);
    GetAudioPids(strl[6], p^.AudioPids);
    GetAC3Pids(strl[6], p^.AC3Pids);
    if AnsiContainsText(strl[7], ';') then
         p^.TTpid := copy(strl[7], 1, AnsiPos(';', strl[7])-1)
    else p^.Ttpid := strl[7];
    p^.SubtPids   := GetSubtitlePids(strl[7]);
    p^.Ca         := strl[8];
    p^.Sid        := strl[9];
    p^.Nid        := strl[10];
    p^.Tid        := strl[11];
    if strl.Count > 12 then
      p^.Rid        := strl[12]
    else p^.Rid := '0';
    if (p^.Source = 'P') OR (p^.Source = 'I') then p^.IPTVParams := strl[2]
    else
    begin
      p^.Bandwidth        := GetParam(strl[2], 'B');
      p^.CodeRH           := GetParam(strl[2], 'C');
      p^.CodeRL           := GetParam(strl[2], 'D');
      p^.Guard            := GetParam(strl[2], 'G');
      p^.Inversion        := GetParam(strl[2], 'I');
      p^.Modulation       := GetParam(strl[2], 'M');
      p^.PilotMode        := GetParam(strl[2], 'N');
      p^.RollOff          := GetParam(strl[2], 'O');
      p^.T2SysID          := GetParam(strl[2], 'Q');
      p^.StreamID         := GetParam(strl[2], 'P');
      p^.System           := GetParam(strl[2], 'S');
      p^.TransmissionMode := GetParam(strl[2], 'T');
      p^.SiMiMode         := GetParam(strl[2], 'X');
      p^.Hierarchy        := GetParam(strl[2], 'Y');
      try
        if (p^.Source[1] = 'S') AND (p^.System = '') then p^.System := '0';
      except
      end;
    end;
  end;
  strl.Free;
end;


procedure InsertGroupSep(str: String);
var
  n, tn: TTreeNode;
  i, j: Integer;
begin
  n := frmMain.tvMain.Items.InsertObject(frmMain.tvMain.Selected, str, NewGroupSep(str));
  frmMain.tvMain.Selected := n;
  frmMain.LastTargetNode := n;
  RenumberChannels;
  if frmMain.pcMain.ActivePage = frmMain.tsInfo then
  begin
    frmMain.edName.SetFocus;
    frmMain.edName.SelectAll;
  end;
end;


function InsertGroupSep(t: TTreeNode; str: String): TTreeNode;
var
  n, tn: TTreeNode;
  i, j: Integer;
begin
  result := NIL;
  n := frmMain.tvMain.Items.InsertObject(t, str, NewGroupSep(str));
  result := n;
end;

procedure InsertChannel(str: String; n: TTreeNode; update: Boolean);
var
  name: String;
begin
  name := copy(str, 1, AnsiPos(':', str)-1);
  name := AnsiReplaceStr(name, '|', ':');
  if isGroupSep(n) then
    n := frmMain.tvMain.Items.AddChildObject(n, name, NewChannel(str))
  else
    n := frmMain.tvMain.Items.InsertObject(n, name, NewChannel(str));
  AssignNodeImage(n);
  if update then
  begin
    RenumberChannels;
    FileChanged(true);
    frmMain.tvMain.Selected := n;
    if frmMain.pcMain.ActivePage = frmMain.tsInfo then
    begin
      frmMain.edName.SetFocus;
      frmMain.edName.SelectAll;
    end;
  end;
end;

procedure InsertNewChannel(str: String; n: TTreeNode);
var
  source, pol: String;
begin
  source := '';
  if isGroupSep(n) AND (n.HasChildren) then
  begin
    source := cleP(n.Items[0].Data)^.Source;
    n := frmMain.tvMain.Items.AddChildObject(n, 'New Channel', NewChannel(str));
  end
  else if (isGroupSep(n)) AND (NOT n.HasChildren) then n := frmMain.tvMain.Items.AddChildObject(n, 'New Channel', NewChannel(str))
  else
  begin
    source := cleP(n.Data)^.Source;
    pol    := cleP(n.Data)^.Pol;
    n := frmMain.tvMain.Items.InsertObject(n, 'New Channel', NewChannel(str));
  end;
  cleP(n.Data)^.Source := source;
  cleP(n.Data)^.Pol    := pol;
  RenumberChannels;
  FileChanged(true);
  frmMain.tvMain.Selected := n;
  if frmMain.pcMain.ActivePage = frmMain.tsInfo then
  begin
    frmMain.edName.SetFocus;
    frmMain.edName.SelectAll;
  end;
end;


// ********* String Helpers ***********//

function LongChannelName(str: String): String;
begin
  if AnsiContainsText(str, ';') then str := copy(str, 1, AnsiPos(';', str)-1);
  if AnsiContainsText(str, ',') then str := copy(str, 1, RPos(',', str)-1);
  result := str;
end;

function ShortChannelName(str: String): String;
begin
  if AnsiContainsText(str, ';') then str := copy(str, 1, AnsiPos(';', str)-1);
  if AnsiContainsText(str, ',') then str := copy(str, RPos(',', str)+1, length(str))
  else  str := '';
  result := AnsiReplaceStr(str, '.', ',');
end;

function ProviderName(str: String): String;
begin
  if AnsiContainsText(str, ';') then result := copy(str, AnsiPos(';', str)+1, length(str))
  else result := '';
end;


// ********** Saving ***************//

function GetProviderStr(n: TTreeNode): String;
begin
  result := ':' + AnsiReplaceStr(n.Text, ':', '|');
end;


function GetParametrStr(n: TTreeNode): String;
  function isPolStr(s: String): Boolean;
  var
    i: Integer;
  begin
    result := true;
    if s = '' then exit;
    for i := 0 to PolMap.Count-1 do
      if UpperCase(s[1]) = PolMap.Names[i] then exit;
    result := false;
  end;
begin
  result := '';
  with cleP(n.Data)^do
  begin
    if Source = '' then Exit;
    if Source[1] in ['S']             then if isPolStr(Pol)          then result += Pol;
    if Source[1] in ['T']             then if Bandwidth <> ''        then result += 'B' + Bandwidth;
    if Source[1] in ['C','S','T']     then if CodeRH <> ''           then result += 'C' + CodeRH;
    if Source[1] in ['T']             then if CodeRL <> ''           then result += 'D' + CodeRL;
    if Source[1] in ['T']             then if Guard <> ''            then result += 'G' + Guard;
    if Source[1] in ['A','C','S','T'] then if Inversion <> ''        then result += 'I' + Inversion;
    if Source[1] in ['A','C','S','T'] then if Modulation <> ''       then result += 'M' + Modulation;
    if Source[1] in ['S']             then if PilotMode <> ''        then result += 'N' + PilotMode;
    if Source[1] in ['S']             then if RollOff <> ''          then result += 'O' + RollOff;
    if Source[1] in ['S','T']         then if StreamID <> ''         then result += 'P' + StreamID;
    if Source[1] in ['T']             then if T2SysID <> ''          then result += 'Q' + T2SysID;
    if Source[1] in ['S','T']         then if System <> ''           then result += 'S' + System;
    if Source[1] in ['T']             then if TransmissionMode <> '' then result += 'T' + TransmissionMode;
    if Source[1] in ['T']             then if SiMiMode <> ''         then result += 'X' + SiMiMode;
    if Source[1] in ['T']             then if Hierarchy <> ''        then result += 'Y' + Hierarchy;
    if Source[1] in ['P','I']         then result := IPTVParams;
  end;
end;

function GetOldParametrStr(n: TTreeNode): String;
  function isPolStr(s: String): Boolean;
  var
    i: Integer;
  begin
    result := true;
    if s = '' then exit;
    for i := 0 to PolMap.Count-1 do
      if s[1] = PolMap.Names[i] then exit;
    result := false;
  end;
begin
  result := '';
  with cleP(n.Data)^do
  begin
    if Source = '' then Exit;
    if Source[1] in ['S']         then if isPolStr(Pol)          then result += Pol;
    if Source[1] in ['C','S','T'] then if Inversion <> ''        then result += 'I' + Inversion;
    if Source[1] in ['C','S','T'] then if CodeRH <> ''           then result += 'C' + CodeRH;
    if Source[1] in ['T']         then if CodeRL <> ''           then result += 'D' + CodeRL;
    if Source[1] in ['C','T']     then if Modulation <> ''       then result += 'M' + Modulation;
    if Source[1] in ['T']         then if Bandwidth <> ''        then result += 'B' + Bandwidth;
    if Source[1] in ['T']         then if TransmissionMode <> '' then result += 'T' + TransmissionMode;
    if Source[1] in ['T']         then if Guard <> ''            then result += 'G' + Guard;
    if Source[1] in ['T']         then if Hierarchy <> ''        then result += 'Y' + Hierarchy;
    if Source[1] in ['P','I']     then result := IPTVParams;
  end;
end;


function GetChannelStr(n: TTreeNode; old: Boolean=false): String;
var
  i: Integer;
begin
  result     := AnsiReplaceStr(cleP(n.data)^.Name, ':', '|');
  if cleP(n.data)^.Short <>'' then
  begin
    cleP(n.data)^.Short := AnsiReplaceStr(cleP(n.data)^.Short, ':', '|');
    cleP(n.data)^.Short := AnsiReplaceStr(cleP(n.data)^.Short, ',', '.');
    result := result + ',' + cleP(n.data)^.Short;
  end
  else if AnsiContainsText(cleP(n.Data)^.Name, ',') then result += ',';
  if cleP(n.data)^.Provider <>'' then
  result := result + ';' + AnsiReplaceStr(cleP(n.data)^.Provider, ':', '|');
  result     := result + ':';
  result     := result + cleP(n.data)^.Frequency + ':';
  if old then result := result + GetOldParametrStr(n) + ':'
  else result := result + GetParametrStr(n) + ':';
  result     := result + cleP(n.data)^.Source + ':';
  result     := result + cleP(n.data)^.Symbolrate + ':';
  result     := result + cleP(n.data)^.Videopid;
  if (cleP(n.data)^.PcrPid <> '0') AND (cleP(n.data)^.PcrPid <> '') AND
     (cleP(n.data)^.PcrPid <> cleP(n.data)^.Videopid) then
                result += '+' + cleP(n.data)^.PcrPid;
  if (cleP(n.data)^.VideoType <> '0') AND (cleP(n.data)^.VideoType <> '') then
                result += '=' + cleP(n.data)^.VideoType;
  result += ':';
  result     := result + AudioPidsToStr(n) + ':';
  result     := result + cleP(n.data)^.TTpid;
  if (cleP(n.Data)^.SubtPids <> NIL) AND (cleP(n.Data)^.SubtPids.Count > 0) then
  begin
    result += ';';
    for i := 0 to cleP(n.Data)^.SubtPids.Count -1 do
    begin
      if i > 0 then result += ',';
      result += cleP(n.Data)^.SubtPids[i];
    end;
  end;
  result     += ':';
  result     := result + cleP(n.data)^.Ca + ':';
  result     := result + cleP(n.data)^.Sid + ':';
  result     := result + cleP(n.data)^.Nid + ':';
  result     := result + cleP(n.data)^.Tid + ':';
  result     := result + cleP(n.data)^.Rid;
end;


function AudioPidsToStr(n: TTreeNode): String;
var
  i: Integer;
begin
  result := '';
  if (cleP(n.Data)^.AudioPids <> NIL) AND (Length(cleP(n.Data)^.AudioPids) > 0) then
  begin
    for i := 0 to Length(cleP(n.Data)^.AudioPids) - 1 do
    begin
      result += cleP(n.Data)^.AudioPids[i, 0];
      if cleP(n.Data)^.AudioPids[i, 1] <> '' then result := result + '=' + cleP(n.Data)^.AudioPids[i, 1];
      if cleP(n.Data)^.AudioPids[i, 2] <> '' then
      begin
        if cleP(n.Data)^.AudioPids[i, 1] = '' then result += '=';
        result := result + '@' + cleP(n.Data)^.AudioPids[i, 2];
      end;
    result += ',';
    end;
    if length(result) > 0 then result := copy(result, 1, length(result)-1);
  end;
  if (cleP(n.Data)^.AC3Pids <> NIL) AND (Length(cleP(n.Data)^.AC3Pids) > 0) then
  begin
    result += ';';
    for i := 0 to Length(cleP(n.Data)^.AC3Pids) - 1 do
    begin
      result += cleP(n.Data)^.AC3Pids[i, 0];
      if cleP(n.Data)^.AC3Pids[i, 1] <> '' then result := result + '=' + cleP(n.Data)^.AC3Pids[i, 1];
      if cleP(n.Data)^.AC3Pids[i, 2] <> '' then
      begin
        if cleP(n.Data)^.AC3Pids[i, 1] = '' then result += '=';
        result := result + '@' + cleP(n.Data)^.AC3Pids[i, 2];
      end;
      result += ',';
    end;
    result := RTrunc(result, 1);
  end;
end;

// ******** Channel properties (TreeNode) **********//

function isTV(n: TTreeNode): Boolean;
begin
  result := (NOT isGroupSep(n)) AND (cleP(n.Data)^.VideoPid <> '0');
end;

function isRadio(n: TTreeNode): Boolean;
begin
  result := (NOT isGroupSep(n)) AND (cleP(n.Data)^.VideoPid <= '1') AND (HasAudio(n));
end;

function isUnknown(n: TTreeNode): Boolean;
begin
  result := (NOT isGroupSep(n)) AND (NOT isTV(n)) AND (NOT isRadio(n));
end;

function isGroupSep(n: TTreeNode): Boolean;
begin
  result := cleP(n.Data)^.GroupSep;
end;

function isIPTV(n: TTreeNode): Boolean;
begin
  result := (cleP(n.Data)^.Source = 'P') OR (cleP(n.Data)^.Source = 'I');
end;

function isAnalog(n: TTreeNode): Boolean;
begin
  result := cleP(n.Data)^.Ca = 'A0';
end;

function isEncrypted(n: TTreeNode): Boolean;
begin
  result := (cleP(n.Data)^.Ca <> '0') AND (cleP(n.Data)^.Ca <> '') AND (cleP(n.Data)^.Ca <> 'A0')
end;

function isObsolete(n: TTreeNode): Boolean;
begin
  result := (AnsiContainsStr(cleP(n.Data)^.Name, 'OBSOLETE')) AND (AnsiContainsStr(cleP(n.Data)^.Provider, 'OBSOLETE'))
end;

function ChannelType(n: TTreeNode): String;
begin
  result := '';
  if isGroupSep(n) then result := 'Group'
  else if isTV(n) then result := 'Television'
  else if isRadio(n) then result := 'Radio'
  else if isUnknown(n) then result := 'Data'
  else if isIPTV(n) then result := 'IPTV'
  else if isAnalog(n) then result := 'Analog'
end;


function ChanIDFreq(pol, freq: String): String;
var
  dummy: Integer;
begin
  result := freq;
  if (length(pol) > 0) AND (UpperCase(pol)[1] in ['H', 'V', 'L', 'R']) then
  begin
    Try
      dummy := StrToInt(freq);
    Except on eConvertError do exit;
    end;
    if UpperCase(pol[1]) = 'H' then dummy += 100000
    else if UpperCase(pol[1]) = 'V' then dummy += 200000
    else if UpperCase(pol[1]) = 'L' then dummy += 300000
    else if UpperCase(pol[1]) = 'R' then dummy += 400000;
    result := IntToStr(dummy);
  end;
end;

function ChannelID(n: TTreeNode): String;
begin
  result := '';
  result += clep(n.Data)^.Source + '-';
  result += clep(n.Data)^.NID + '-';
  if (cleP(n.Data)^.Tid = '0') AND (cleP(n.Data)^.NID = '0') then
    result += ChanIdFreq(clep(n.Data)^.Pol, clep(n.Data)^.Frequency) + '-'
  else result += clep(n.Data)^.Tid + '-';
  result += clep(n.Data)^.SID + '-';
  result += clep(n.Data)^.Rid;
end;

function LanguagesToStr(n: TTreeNode): String;
var
  i: Integer;
  dummy: String;
begin
  result := '';
  dummy := '';
  if (cleP(n.Data)^.AudioPids = NIL) OR (Length(cleP(n.Data)^.AudioPids) = 0) then Exit;
  for i := 0 to Length(cleP(n.Data)^.AudioPids)-1 do
  begin
    if cleP(n.Data)^.AudioPids[i,0] = '0' then continue;
    dummy := cleP(n.Data)^.AudioPids[i, 1];
    if dummy <> '' then result += dummy + ', ';
  end;
  result := RTrunc(result, 2);
end;

function HasDolby(n: TTreeNode): Boolean;
begin
  if (cleP(n.Data)^.AC3Pids <> NIL) AND (Length(cleP(n.Data)^.AC3Pids) > 0) then result := true
  else result := false;
end;

function HasAudio(n: TTreeNode): Boolean;
begin
  if (cleP(n.Data)^.AudioPids <> NIL) AND (Length(cleP(n.Data)^.AudioPids) > 0) AND
     (cleP(n.Data)^.AudioPids[0, 0] <> '0') then result := true
  else result := false;
end;

function HasMultiLang(n: TTreeNode): Boolean;
begin
  if (cleP(n.Data)^.AudioPids <> NIL) AND (Length(cleP(n.Data)^.AudioPids) > 1) then result := true
  else result := false;
end;



// ******** Channel properties (String) **********//


function GetAudioPid(str: String): String;
begin
  if AnsiContainsText(str, '=') then
     result := copy(str, 1, AnsiPos('=', str) - 1)
  else result := str;
end;

function GetLanguage(str: String): String;
begin
  if AnsiContainsText(str, '=') then
  begin
     result := copy(str, AnsiPos('=', str) + 1, length(str));
     if AnsiContainsText(result, '@') then result := copy(result, 1, AnsiPos('@', result)-1);
  end
  else result := '';
end;

function GetAudVidType(str: String): String;
begin
  if AnsiContainsText(str, '@') then result := copy(str, AnsiPos('@', str)+1, length(str))
  else result := '';
end;

function GetVidPid(str: String): String;
begin
  if AnsiContainsText(str, '+') then
    result := copy(str, 1, AnsiPos('+', str)-1)
  else result := str;
  if AnsiContainsText(result, '=') then result := copy(result, 1, AnsiPos('=', result)-1);
end;

function GetVidType(str: String): String;
begin
  if AnsiContainsText(str, '=') then
    result := copy(str, AnsiPos('=', str)+1, length(str))
  else result := '0';
end;

function GetPcrPid(str: String): String;
begin
  if str = '0' then result := '';
  if AnsiContainsText(str, '+') then
    result := copy(str, AnsiPos('+', str)+1, length(str))
  else result := '0';
  if AnsiContainsText(result, '=') then result := copy(result, 1, AnsiPos('=', result)-1);
end;

function GetSubtitlePids(str: String): TStringList;
var
  dummy: String;
  SubList: TStringList;
begin
  if AnsiContainsText(str, ';') then
  begin
    dummy := copy(str, AnsiPos(';', str)+1,length(str));
    result := StringSplit(dummy, ',');
  end
  else result := TStringList.Create;
end;

function GetParam(str: String; c: Char): String;
var
  i: Integer;
begin
  result := '';
  i := Pos(c, str);
  if i > 0 then
  begin
    While (i < Length(str)) AND (str[i+1] in ['0'..'9']) do
    begin
      result += str[i+1];
      inc(i);
    end;
  end;
end;

procedure GetAudioPids(str: String; var AudArray: TAudioArray);
var
  AudList: TStringList;
  i: Integer;
begin
  if AnsiContainsStr(str, ';') then str := copy(str, 1, Pos(';', str) -1);
  AudList := StringSplit(str, ',');
  SetLength(AudArray, AudList.Count);
  for i := 0 to AudList.Count-1 do
  begin
    AudArray[i, 0] := GetAudioPid(AudList[i]);
    AudArray[i, 1] := GetLanguage(AudList[i]);
    AudArray[i, 2] := GetAudVidType(AudList[i]);
  end;
end;

Procedure GetAC3Pids(str: String; var AC3Array: TAudioArray);
var
  AC3List: TStringList;
  i: Integer;
begin
  if AnsiContainsStr(str, ';') then str := copy(str, Pos(';', str) + 1, length(str))
  else
  begin
    SetLength(AC3Array, 0);
    exit;
  end;
  AC3List := StringSplit(str, ',');
  SetLength(AC3Array, AC3List.Count);
  for i := 0 to AC3List.Count-1 do
  begin
    AC3Array[i,0] := GetAudioPid(AC3List[i]);
    AC3Array[i, 1] := GetLanguage(AC3List[i]);
    AC3Array[i, 2] := GetAudVidType(AC3List[i]);

  end;
end;

function PolToStr(str: String): String;
begin
  result := '';
  if length(str) < 1 then Exit;
  Case UpperCase(str)[1] of
          'H': result := 'Horizontal';
          'V': result := 'Vertical';
          'L': result := 'Left Circular';
          'R': result := 'Right Circular';
  end;
end;

function EncryptionsToStr(n: TTreeNode): String;
var
  CaSystems, ca: TStringList;
  i, d: Integer;
begin
  result := '';
  if cleP(n.Data)^.Ca = '0' then Exit;
  CaSystems := StringSplit(cleP(n.Data)^.Ca, ',');
  ca := TStringList.Create;
  ca.Sorted := true;
  ca.Duplicates := dupIgnore;
  for i := 0 to CaSystems.Count - 1 do
  begin
    if NOT TryStrToInt('$'+CaSystems[i], d) then continue;
    if d >= 256 then
    begin
      d := d DIV 256;
      case d of
         1:  ca.Add('Seca/Mediagard');
         5:  ca.Add('Viaccess');
         6:  ca.Add('Irdeto');
         9:  ca.Add('NDS/Videoguard');
        11:  ca.Add('Conax');
        13:  ca.Add('CryptoWorks');
        14:  ca.Add('PowerVu');
        18:  ca.Add('NagraVision');
        23:  ca.Add('BetaCrypt');
        24:  ca.Add('NagraVision');
        74:  ca.Add('SkyCrypt');
      end
    end
    else if (d >= 16) AND (d <= 255) then
    begin
      if d = 160 then ca.Add('TV-analog')
      else ca.Add('User defined')
    end
    else if (d >= 1) AND (d <= 15) then ca.Add('DVB'+ CaSystems[i]);
  end;
  ca.Sort;
  for i := 0 to ca.Count - 1 do result += ca[i] + ', ';
  result := RTrunc(result, 2);
  CaSystems.Free;
  Ca.Free;
end;


// *********** TreeView Helpers ************//

procedure RenumberChannels;
var
  i, j, r: Integer;
begin
  j := 0;
  r := 0;
  For i := 0 to frmMain.tvMain.Items.Count-1 do
  begin
    if (isGroupSep(frmMain.tvMain.Items[i])) AND (isJumpMark(frmMain.tvMain.Items[i]))
            then
            if (GetPrevChannel(frmMain.tvMain.Items[i]) <> NIL) AND
            (cleP(GetPrevChannel(frmMain.tvMain.Items[i]).Data)^.Number < GetJumpIndex(frmMain.tvMain.Items[i])) then
            j := GetJumpIndex(frmMain.tvMain.Items[i])-1;
    if NOT isGroupSep(frmMain.tvMain.Items[i]) then
    begin
      inc(j);
      inc(r);
      cleP(frmMain.tvMain.Items[i].Data)^.Number := j;
    end
    else cleP(frmMain.tvMain.Items[i].Data)^.Number := 0;
  end;
  sbm(GetChannelsInfo, 1);
end;

function GetPrevChannel(n: TTreeNode): TTreeNode;
begin
  result := n;
  While (result <> NIL) AND (isGroupSep(result)) do result := result.GetPrev;
end;

function GetNextChannel(n: TTreeNode): TTreeNode;
begin
  result := n;
  While (result <> NIL) AND (isGroupSep(result)) do result := result.GetNext;
end;

function GetNextGroupSep(n: TTreeNode): TTreeNode;
var
  tn: TTreeNode;
begin
  result := NIL;
  tn := n.GetNext;
  While tn <> NIL do
  begin
    if isGroupSep(tn) then
    begin
      result := tn;
      exit;
    end;
    tn := tn.GetNext;
  end;
end;

function HasWrongGroupSep(n: TTreeNode): Boolean;
var
  i: Integer;
begin
  result := false;
  if isGroupSep(n) then Exit;
  for i := 0 to frmMain.tvMain.Items.Count -1 do
  begin
    if (NOT isGroupSep(frmMain.tvMain.Items[i])) OR (frmMain.tvMain.Items[i].HasChildren = false) then continue;
    if (frmMain.tvMain.Items[i].Text = cleP(n.Data)^.Provider) AND
       (cleP(frmMain.tvMain.Items[i+1].data)^.Source = cleP(n.Data)^.Source) AND
       (ChannelType(frmMain.tvMain.Items[i+1]) = ChannelType(n)) AND
       (n.HasAsParent(frmMain.tvMain.Items[i]) = false) then
    begin
      result := true;
      Exit;
    end;
  end;
end;


function GetRightGroupSep(n: TTreeNode): TTreeNode;
var
  i: Integer;
begin
  result := NIL;
  if isGroupSep(n) then Exit;
  for i := 0 to frmMain.tvMain.Items.Count -1 do
  begin
    if (NOT isGroupSep(frmMain.tvMain.Items[i]))  OR (frmMain.tvMain.Items[i].HasChildren = false) then continue;
    if (frmMain.tvMain.Items[i].Text = cleP(n.Data)^.Provider) AND
       (cleP(frmMain.tvMain.Items[i+1].data)^.Source = cleP(n.Data)^.Source) AND
       (ChannelType(frmMain.tvMain.Items[i+1]) = ChannelType(n)) then
    begin
      result := frmMain.tvMain.Items[i];
      Exit;
    end;
  end;
end;


function GetAllSources(var sl: TStringList): Integer;
var
  i: Integer;
begin
  result := 0;
  if frmMain.tvMain.Items.Count = 0 then exit;
  for i := 0 to frmMain.tvMain.Items.Count-1 do
  begin
    if (isGroupSep(frmMain.tvMain.Items[i])) OR
       (trim(cleP(frmMain.tvMain.Items[i].Data)^.Source) = '') then continue;
    if sl.IndexOf(cleP(frmMain.tvMain.Items[i].Data)^.Source) < 0 then
      sl.Add(cleP(frmMain.tvMain.Items[i].Data)^.Source)
  end;
  result := sl.Count;
end;

procedure RemoveSources(source: String);
var
  i: Integer;
  tp, t: TTreeNode;
begin
  i := frmMain.tvMain.Items.Count - 1;
  While i > 0 do
  begin
    if cleP(frmMain.tvMain.Items[i].Data)^.Source = source then
    begin
      tp := frmMain.tvMain.Items[i].Parent;
      DeleteNode(frmMain.tvMain, frmMain.tvMain.Items[i], false);
      if (tp <> NIL) AND
         (cleP(tp.Data)^.GroupSep = true) AND
         (tp.HasChildren = false) then
         begin
           DeleteNode(frmMain.tvMain, tp, false);
           dec(i);
         end;
    end;
    dec(i);
  end;
  RenumberChannels;
  FileChanged(true);
end;

procedure RemoveAll(target: String);
var
  i: Integer;
  tp, t: TTreeNode;
begin
  i := frmMain.tvMain.Items.Count - 1;
  While i > 0 do
  begin
    if ((target = 'TV') AND (isTV(frmMain.tvMain.Items[i]))) OR
       ((target = 'Radio') AND (isRadio(frmMain.tvMain.Items[i]))) OR
       ((target = 'Data') AND (isUnknown(frmMain.tvMain.Items[i]))) OR
       ((target = 'Encrypted') AND (isEncrypted(frmMain.tvMain.Items[i]))) OR
       ((target = 'Obsolete') AND (isObsolete(frmMain.tvMain.Items[i]))) then
    begin
      tp := frmMain.tvMain.Items[i].Parent;
      DeleteNode(frmMain.tvMain, frmMain.tvMain.Items[i], false);
      if (tp <> NIL) AND
         (cleP(tp.Data)^.GroupSep = true) AND
         (tp.HasChildren = false) then
         begin
           DeleteNode(frmMain.tvMain, tp, false);
           dec(i);
         end;
    end;
    dec(i);
  end;
  RenumberChannels;
  FileChanged(true);
end;

initialization
BuildParameterMaps;

finalization
RemoveParameterMaps;


end.

