
//
// cle4vdr 2005 - 2018 by Alexander Hans <cle4vdr@arcor.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
// Or, point your browser to http://www.gnu.org/copyleft/gpl.html
//


unit uedit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ComCtrls, StdCtrls, StrUtils, Grids, Spin, uchannels;

procedure ChangeChannelParams(Sender: TObject; n: TTreeNode);
function ChangeChannelData(n: TTreeNode; str: String; Sender: TObject): Boolean;

function CopyNode(n: TTreeNode): cleP;
procedure PasteNode(p, t: cleP);
function FindNextToSelect(n: TTreeNode): TTreeNode;
procedure MoveNode(SelNode, ToNode: TTreeNode);
procedure DeleteNode(tv: TTreeView; n: TTreeNode; update: Boolean);
procedure DeleteNodeChilds(tv: TTreeView; n: TTreeNode);

function CopyDataHasGroupSep: Boolean;
procedure ClearCopyData;

function ChangeAudioSubtData(sg: TStringGrid; n: TTreeNode): Boolean;

implementation

uses uinterface, uutils, uinfo;

procedure ChangeChannelParams(Sender: TObject; n: TTreeNode);
begin
  if n = NIL then Exit;
  if Sender.ClassType = TComboBox then
  begin
    if (Sender as TComboBox) = frmMain.cbBandwidth then cleP(n.Data)^.Bandwidth := BandwidthMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbCodeRH then cleP(n.Data)^.CodeRH := CodeRMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbCodeRL then cleP(n.Data)^.CodeRL := CodeRMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbGuard then cleP(n.Data)^.Guard := GuardMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbInversion then cleP(n.Data)^.Inversion := InversionMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbMod then cleP(n.Data)^.Modulation := ModulationMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbPilotM then cleP(n.Data)^.PilotMode := PilotModeMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbRolloff then cleP(n.Data)^.RollOff := RolloffMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbSystem then cleP(n.Data)^.System := SystemMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbTransmission then cleP(n.Data)^.TransmissionMode := TransmissionMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbSiMiMode then cleP(n.Data)^.SiMiMode := SiMiModeMap.Names[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbHierarchy then cleP(n.Data)^.Hierarchy := HierarchyMap.Names[(Sender as TComboBox).ItemIndex]

    else if (Sender as TComboBox) = frmMain.cbPol then cleP(n.Data)^.Pol := PolMap[(Sender as TComboBox).ItemIndex]
    else if (Sender as TComboBox) = frmMain.cbSourceExp then
    begin
      cleP(n.Data)^.Source          := frmMain.cbSourceExp.Caption;
      frmMain.cbSource.ItemIndex    := frmMain.cbSourceExp.ItemIndex;
      frmMain.edSource.Text         := frmMain.cbSource.Text;
    end
    else if (Sender as TComboBox) = frmMain.cbSource then
    begin
      frmMain.cbSourceExp.ItemIndex := frmMain.cbSource.ItemIndex;
      frmMain.edSource.Text         := frmMain.cbSource.Text;
      cleP(n.Data)^.Source          := frmMain.cbSourceExp.Caption;
      frmMain.cbSource.Visible      := false;
    end;
  end
  else if Sender.ClassType = TSpinEdit then
  begin
    if (Sender as TSpinEdit) = frmMain.seT2SysID then cleP(n.Data)^.T2SysID := frmMain.seT2SysID.Text
    else if (Sender as TSpinEdit) = frmMain.seStreamID then cleP(n.Data)^.StreamID := frmMain.seStreamID.Text
  end;


  frmMain.tvMainSelectionChanged(frmMain);
  FileChanged(true);
end;

function ChangeChannelName(new: String; var target: String; n: TTreeNode; isName: Boolean = true): Boolean;
begin
  result := false;
  new := trim(new);
  if (AnsiContainsText(new, ',') OR (AnsiContainsText(new, ';'))) then exit;
  if NOT isGroupSep(n) then new := StripTrailingChar(new, [':', '@'])
  else new := StripTrailingChar(new, [':']);
  result := true;
  if isJumpMark(n) then
  begin
    result := TryJumpMark(new);
    if result = true then RenumberChannels;
  end;
  if result = true then
  begin
    if isName then n.Text := new;
    target := new;
  end;
end;

function ChangeIPTVData(AStr: String; var APolStr: String): Boolean;
begin
  result := false;
  if Pos(':', AStr) < 0 then Exit;
  APolStr := AStr;
  result := true;
end;

function ChangeChannelData(n: TTreeNode; str: String; Sender: TObject): Boolean;
begin
  result := true;
  with cleP(n.Data)^ do
  begin
    with frmMain do
    begin
      if Sender = edName               then result     := ChangeChannelname(str, cleP(n.Data)^.Name, n)
      else if Sender = edNameExp       then result     := ChangeChannelname(str, cleP(n.Data)^.Name, n)
      else if Sender = edShort         then result     := ChangeChannelname(str, Short, n, false)
      else if Sender = edShortExp      then result     := ChangeChannelname(str, Short, n, false)
      else if Sender = edProvider      then result     := ChangeChannelname(str, Provider, n, false)
      else if Sender = edProviderExp   then result     := ChangeChannelname(str, Provider, n, false)
      else if Sender = edSource        then result     := TrySource(str, Source)
      else if Sender = edFrequencyExp  then result     := TryInteger(str, Frequency)
      else if Sender = edSymbolrateExp then result     := TryInteger(str, Symbolrate)
  //    else if Sender = edSourceExp     then result     := TrySource(str, Source)
      else if Sender = edVid           then result     := TryInteger(str, Videopid)
      else if Sender = edPCR           then result     := TryInteger(str, PcrPid)
      else if Sender = edTeletext      then result     := TryInteger(str, Ttpid)
      else if Sender = edTid           then result     := TryInteger(str, Tid)
      else if Sender = edSid           then result     := TryInteger(str, Sid)
      else if Sender = edNit           then result     := TryInteger(str, Nid)
      else if Sender = edRad           then result     := TryInteger(str, Rid)
      else if Sender = edCa            then result     := TryCa(str, Ca)
      else if Sender = edIPTVParams    then result     := ChangeIPTVData(str, Pol);
    end;
  end;
end;

function CopyNode(n: TTreeNode): cleP;
begin
  New(result);
  result^ := cleP(n.Data)^;
end;

procedure PasteNode(p, t: cleP);
begin
  t^ := p^;
  FileChanged(true);
end;

function FindNextToSelect(n: TTreeNode): TTreeNode;
begin
  if n.Level = 0 then
  begin
    if n.Index <> 0 then result := n.GetPrevSibling
    else result := n.GetNextSibling;
  end
  else
  begin
    if (n = n.Parent.GetFirstChild) AND (n.Parent.Count > 1) then result := n.GetNext
    else result := n.GetPrev;
  end;
end;

procedure MoveNode(SelNode, ToNode: TTreeNode);
begin
  frmMain.SaveTreeState;
  SelNode.MoveTo(ToNode, naAddChild);
  frmMain.RestoreTreeState;
  RenumberChannels;
  FileChanged(true);
end;

procedure DeleteNode(tv: TTreeView; n: TTreeNode; update: Boolean);
var
  s: TTreeNode;
  i: Integer;
begin
  s := NIL;
  if n = NIL then Exit;
  s := FindNextToSelect(n);
  if n.HasChildren then
  begin
    for i := 0 to n.Count -1 do Dispose(cleP(n[i].Data));
    n.DeleteChildren;
  end;
  Dispose(cleP(n.Data));
  n.Delete;
  if update then
  begin
    if tv.Items.Count > 0 then
    begin
      if s <> NIL then
        tv.Selected := s
      else tv.Selected := tv.Items[0];
    end;
    RenumberChannels;
    FileChanged(true);
  end;
end;

procedure DeleteNodeChilds(tv: TTreeView; n: TTreeNode);
var
  s: TTreeNode;
  i: Integer;
begin
  s := NIL;
  if n = NIL then Exit;
  s := FindNextToSelect(n);
  if n.HasChildren then
  begin
    for i := 0 to n.Count -1 do Dispose(cleP(n[i].Data));
    n.DeleteChildren;
  end;
  if tv.Items.Count > 0 then
  begin
    if s <> NIL then
      tv.Selected := s
    else tv.Selected := tv.Items[0];
  end;
  RenumberChannels;
  FileChanged(true);
end;


function CopyDataHasGroupSep: Boolean;
var
  i: Integer;
begin
  result := false;
  if length(CopyData) = 0 then Exit;
  for i := 0 to length(CopyData)-1 do
  begin
    if CopyData[i]^.Source = '' then
    begin
      result := true;
      Exit;
    end;
  end;
end;

procedure ClearCopyData;
var
  i: Integer;
begin
  if Length(CopyData) > 0 then
    for i := 0 to length(CopyData)-1 do
    begin
      Dispose(CopyData[i]);
      CopyData[i] := NIL;
    end;
  setlength(CopyData, 0);
end;

function ChangeAudioSubtData(sg: TStringGrid; n: TTreeNode): Boolean;
var
  i: Integer;
begin
  if sg = frmMain.sgAudio then
  begin
    SetLength(cleP(n.Data)^.AudioPids, 0);
    for i := 1 to sg.RowCount -1 do
    begin
      SetLength(cleP(n.Data)^.AudioPids, Length(cleP(n.Data)^.AudioPids)+1);
      cleP(n.Data)^.AudioPids[i-1, 0] := sg.Cells[0, i];
      cleP(n.Data)^.AudioPids[i-1, 1] := sg.Cells[1, i];
      cleP(n.Data)^.AudioPids[i-1, 2] := sg.Cells[2, i];
    end;
  end
  else if sg = frmMain.sgDolby then
  begin
    SetLength(cleP(n.Data)^.AC3Pids, 0);
    for i := 1 to sg.RowCount -1 do
    begin
      SetLength(cleP(n.Data)^.AC3Pids, Length(cleP(n.Data)^.AC3Pids)+1);
      cleP(n.Data)^.AC3Pids[i-1, 0] := sg.Cells[0, i];
      cleP(n.Data)^.AC3Pids[i-1, 1] := sg.Cells[1, i];
      cleP(n.Data)^.AC3Pids[i-1, 2] := sg.Cells[2, i];
    end;
  end
  else if sg = frmMain.sgSubttl then
  begin
    cleP(n.Data)^.SubtPids.Clear;
    for i := 1 to sg.RowCount -1 do
    begin
      if sg.Cells[1, i] <> '' then
        cleP(n.Data)^.SubtPids.Add(sg.Cells[0, i] + '=' + sg.Cells[1, i])
      else  cleP(n.Data)^.SubtPids.Add(sg.Cells[0, i]);
    end;
  end;
  result := true;
end;

end.

