
//
// cle4vdr 2005 - 2018 by Alexander Hans <cle4vdr@arcor.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
// Or, point your browser to http://www.gnu.org/copyleft/gpl.html
//


unit uiodvb;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Dialogs, StrUtils, ComCtrls;

Type
    TDVBRec = record
    Sign : array [1..4] of Char;
    ServID : Word;
    PmtPID : Word;
    Freq : Word;
    Symbrt : Word;
    FEC : Byte;
    AcqBW : Byte;
    Pol : Byte;
    diseqc : Byte;
    VidPID : Word;
    AudPID : Word;
    PcrPID : Word;
    Ac3PID : Word;
    EcmPID : Word;
    VidFlag : Byte;
    ServiceType : Byte;
    ttxPID : Word;
    Transp : Word;
    Name : array[1..24] of Char;
    Flags : Byte;
    ProvIndex : Byte;
    ParentLock : Byte;
    CountryCode : Byte;
    ChannelLink : Byte;
    digicipher : Byte;
    NetzwID : Word;
end;

type
  NameArray = array [1..24] of Char;
  
const
  DVBFecMap: array[1..6] of String =
             ('999', '12', '23', '34', '56', '78');
             
procedure WriteDVB(const path: String);

implementation

uses uinterface, uchannels, uutils;

function DVBChannelName(name, short: String): NameArray;
var
  i, j: Integer;
  sl: TStringList;
  dummy: String;
begin
  if length(name) > 24 then name := copy(name, 1, 24);
  for i := 1 to length(result) do result[i] := #0;
  if short = '' then
    for i := 1 to length(name) do result[i] := name[i]
  else
  begin
    sl := StringSplit(short, ' ');
    dummy := name;
    for j := 0 to sl.Count - 1 do
    begin
      i := Pos(sl[j], dummy);
      if i > 0 then
      begin
        dummy := InsertText(dummy, #134, i-1);
        dummy := InsertText(dummy, #135, i+length(sl[j]));
      end
      else result := name;
    end;
  if length(dummy) > 24 then dummy := copy(dummy, 1, 24);
  for i := 1 to length(dummy) do result[i] := dummy[i];
  sl.Free;
  end;
end;

function GetOrbitPos(n: TTreeNode): Word;
var
  dummy: String;
  i: Integer;
begin
  result := 0;
  n := GetNextChannel(n);
  if (n <> NIL) AND (StartsWith(cleP(n.Data)^.Source, 'S')) then
  begin
    dummy := cleP(n.Data)^.Source;
    dummy := LTrunc(dummy, 1);
    dummy := RTrunc(dummy, 1);
    dummy := AnsiReplaceText(dummy, '.', '');
    if TryStrToInt(dummy, i) then result := Word(i)
  end;
end;

function SourceToLnb(source: String; SourceList: TStringList): Integer;
begin
  result := SourceList.IndexOf(source);
  if result < 0 then
  begin
    SourceList.Add(source);
    result := SourceList.IndexOf(source);
  end;
  if result > 4 then result := 4;
end;

function CodeRToFec(cr: String): Integer;
begin
  if cr = '12' then result := 2
  else if cr = '23' then result := 3
  else if cr = '34' then result := 4
  else if cr = '56' then result := 6
  else if cr = '78' then result := 7
  else result := 0;
end;

function CheckDVBInt(chk: String; ff: Boolean = false): Word;
var
  i: Integer;
begin
  if (chk = '') OR (chk = '0') OR (NOT TryStrToInt(chk, i)) then
  begin
    if ff then result := 8191
    else result := 0;
  end
  else result := Word(StrToInt(chk));
end;

procedure WriteDVB(const path: String);
var
  outf: File;
  OutRec: TDVBRec;
  i: Integer;
  sources: TStringList;
begin
  AssignFile(outf, path);
  Rewrite(outf, 1);
  Sources:= TStringList.Create;
  For i := 0 to frmMain.tvMain.Items.Count - 1 do
  begin
    try
    OutRec.Sign   := 'DVSO';
    OutRec.Name   := DVBChannelName(cleP(frmMain.tvMain.Items[i].Data)^.Name, cleP(frmMain.tvMain.Items[i].Data)^.Short);
    if isGroupSep(frmMain.tvMain.Items[i]) then
    begin
      OutRec.Freq   := 0;
      OutRec.PcrPID := Swap(GetOrbitPos(frmMain.tvMain.Items[i]));
      OutRec.ServID      := 0;
      OutRec.PmtPID      := 0;
      OutRec.Symbrt      := 0;
      OutRec.FEC         := 0;
      OutRec.AcqBW       := 0;
      OutRec.Pol         := 0;
      OutRec.diseqc      := 0;
      OutRec.VidPID      := 0;
      OutRec.AudPID      := 0;
      OutRec.Ac3PID      := 0;
      OutRec.EcmPID      := 0;
      OutRec.VidFlag     := 0;
      OutRec.ServiceType := 0;
      OutRec.ttxPID      := 0;
      OutRec.Transp      := 0;
      OutRec.Flags       := 0;
      OutRec.ProvIndex   := 0;
      OutRec.ParentLock  := 0;
      OutRec.CountryCode := 0;
      OutRec.ChannelLink := 0;
      OutRec.digicipher  := 0;
      OutRec.NetzwID     := 0;
    end
    else
    begin
      OutRec.ServID := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.Sid));
      OutRec.PmtPID := 0;
      OutRec.Freq   := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.Frequency));
      OutRec.Symbrt := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.Symbolrate));
      OutRec.FEC    := CodeRToFec(cleP(frmMain.tvMain.Items[i].Data)^.CodeRH);
      OutRec.AcqBW  := 0;  // ?
      if UpperCase(cleP(frmMain.tvMain.Items[i].Data)^.Pol) = 'V' then
        OutRec.Pol  := 1
      else OutRec.Pol := 0;
      OutRec.diseqc := SourceToLnb(cleP(frmMain.tvMain.Items[i].Data)^.Source, Sources);
      if cleP(frmMain.tvMain.Items[i].Data)^.Inversion = '1' then OutRec.diseqc += 8
      else if cleP(frmMain.tvMain.Items[i].Data)^.Inversion = '999' then OutRec.diseqc += 9;
      OutRec.VidPID := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.Videopid, true));
      OutRec.AudPID := SwapEndian(CheckDVBInt(GetAudioPid(cleP(frmMain.tvMain.Items[i].Data)^.AudioPids[0,0])));
      OutRec.PcrPID := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.PcrPid));
      if HasDolby(frmMain.tvMain.Items[i]) then
        OutRec.Ac3PID := SwapEndian(CheckDVBInt(GetAudioPid(cleP(frmMain.tvMain.Items[i].Data)^.AC3Pids[0,0])))
      else
        OutRec.Ac3PID := 0;
      OutRec.EcmPID := 0;
      OutRec.VidFlag:= 0; // 0 PAl; 1 NTSC
      if isTV(frmMain.tvMain.Items[i]) then
        OutRec.ServiceType := 1
      else if isRadio(frmMain.tvMain.Items[i]) then
        OutRec.ServiceType := 2
      else OutRec.ServiceType := 0;
      OutRec.ttxPID := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.Ttpid, true));
      OutRec.Transp := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.Tid));
      OutRec.Flags  := 0; // AutoPMT  + 128 FreeCa
      if isEncrypted(frmMain.tvMain.Items[i]) then OutRec.Flags  += 128;
      OutRec.ProvIndex := 0;
      OutRec.ParentLock := 0;
      OutRec.CountryCode := 0;
      OutRec.ChannelLink := 0;
      OutRec.digicipher := 0;
      OutRec.NetzwID := SwapEndian(CheckDVBInt(cleP(frmMain.tvMain.Items[i].Data)^.Nid));
    end;
    Except
      Showmessage(inttostr(i) + ' Error in: ' +cleP(frmMain.tvMain.Items[i].Data)^.Name);
    end;
    Blockwrite(outf, OutRec, SizeOf(TDVBRec));
  end;
  CloseFile(outf);
  Sources.Free;
end;

end.

